from collections import defaultdict
import logging
from typing import List, Optional, Dict
import time

import win32pdh

default_logger = logging.getLogger(__name__)


class Counter:
    # TODO: This needs to be split into Metric, Instance, Collector
    # TODO: Collector has the key and aggregation
    # TODO: Metric has the Object and Counter, and List of Instance
    def __init__(
        self,
        obj: str,
        counter: str,
        instance: Optional[str],
        handle,
        key: str,
        metric_type: str,
        aggregation: str,
        dimension_name: str = None,
    ):
        self.path = f"{obj}({instance})/{counter}" if instance is not None else f"{obj}/{counter}"
        self.obj = obj
        self.instance = instance
        self.counter = counter
        self.handle = handle
        self.key = key
        self.metric_type = metric_type
        self.value = 0
        self.aggregation = aggregation
        self.dimension_name = dimension_name

    def __str__(self):
        return f"Counter({self.path}, {self.handle}, {self.key}, value: {self.value})"

    def __repr__(self):
        return self.__str__()


class Perfmon:
    def __init__(self, logger: logging.Logger = default_logger):
        self.logger = logger
        self.counters: List[Counter] = []

    def __enter__(self):
        self.query = win32pdh.OpenQuery()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.query is not None:
            win32pdh.CloseQuery(self.query)

    def get_counters_per_key(self) -> Dict[str, List[Counter]]:
        metrics_dict: Dict[str, List[Counter]] = defaultdict(list)

        for counter in self.counters:
            metrics_dict[counter.key].append(counter)

        return metrics_dict

    def _get_all_instances(self, obj) -> List[str]:

        try:
            _, instances = win32pdh.EnumObjectItems(None, None, obj, win32pdh.PERF_DETAIL_WIZARD)
        except Exception as e:
            return ["*"]

        return instances

    def _add_counter_path(
        self, obj: str, counter: str, instance: Optional[str], key: str, aggregation, metric_type="absolute", dimension_name=None
    ):
        if instance is None and aggregation == "count":
            self.counters.append(
                Counter(obj, counter, instance, None, key, metric_type, aggregation, dimension_name=dimension_name)
            )
        else:
            path = win32pdh.MakeCounterPath((None, obj, instance, None, -1, counter))
            counter_handle = win32pdh.AddEnglishCounter(self.query, path)
            self.counters.append(
                Counter(obj, counter, instance, counter_handle, key, metric_type, aggregation, dimension_name=dimension_name)
            )

    def add_counter(self, obj: str, counter: str, key="", aggregation="avg", metric_type="absolute", dimension_name=None):
        try:
            instances = self._get_all_instances(obj)
            self.logger.debug(f"Attempting to add counter, object='{obj}', counter='{counter}, instances='{instances}'")
            for instance in instances:
                self._add_counter_path(obj, counter, instance, key, aggregation, metric_type, dimension_name=dimension_name)
            if not instances:
                self._add_counter_path(obj, counter, None, key, aggregation, metric_type, dimension_name=dimension_name)

        except Exception as e:
            self.logger.debug(f"Could not add counter: '{obj}\\{counter}': {e}")

    def collect_data(self):
        try:
            win32pdh.CollectQueryData(self.query)
            time.sleep(1)
            win32pdh.CollectQueryData(self.query)

            for counter in self.counters:
                try:
                    if counter.handle is not None:
                        counter.value = win32pdh.GetFormattedCounterValue(counter.handle, win32pdh.PDH_FMT_DOUBLE)[1]
                except Exception as e:
                    self.logger.error(f"Could not obtain value for counter {counter}: {e}")
        except Exception as e:
            self.logger.error(f"Could not call CollectQueryData: {e}")


def main():
    logger = logging.getLogger(__name__)
    st = logging.StreamHandler()
    logger.setLevel(logging.DEBUG)
    fmt = logging.Formatter("%(asctime)s - %(levelname)s - %(funcName)s - %(message)s ")
    st.setFormatter(fmt)
    logger.addHandler(st)

    with Perfmon(logger) as p:
        p.add_counter("MSMQ Queue", "Messages in Queue", "msgs")
        p.add_counter("Memory", "Available Bytes", "mem")
        p.add_counter("ICA Session", "Latency - Last Recorded", "mem")
        p.collect_data()

        for metric, counters in p.get_counters_per_key().items():
            print(metric, sum([counter.value for counter in counters]))


if __name__ == "__main__":
    main()
